#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "han_lib.h"

int first_nonopt;
int last_nonopt;
int __getopt_initialized;
int optind;
char *optarg, *nextchar;
char * posixly_correct;
int opterr;
int optopt;
enum {
	REQUIRE_ORDER, PERMUTE, RETURN_IN_ORDER
} ordering;

#define NONOPTION_P (argv[optind][0] != '-' || argv[optind][1] == '\0')

# ifdef USE_NONOPTION_FLAGS
#  define SWAP_FLAGS(ch1, ch2) \
   if (nonoption_flags_len > 0)                                                \
     {                                                                         \
       char __tmp = __getopt_nonoption_flags[ch1];                             \
       __getopt_nonoption_flags[ch1] = __getopt_nonoption_flags[ch2];          \
       __getopt_nonoption_flags[ch2] = __tmp;                                  \
   }
# else
#  define SWAP_FLAGS(ch1, ch2)
# endif


struct option {
	char * name;
	int has_arg;
	int *flag;
	int val;
};

void
exchange(char ** argv)
{
	int bottom = first_nonopt;
	int middle = last_nonopt;
	int top = optind;
	char * tem;

	while(top > middle && middle > bottom) {
		if(top - middle > middle - bottom) {
			int len = middle - bottom;
			int i;

			for(i=0; i<len;i++) {
				tem = argv[bottom+i];
				argv[bottom+i]=argv[top-(middle-bottom)+i];
				argv[top-(middle-bottom)+i]=tem;
				SWAP_FLAGS(bottom+i,top-(middle-bottom)+i);
			}
			top -= len;
		} else {
			int len = top-middle;
			int i;
			for(i=0;i<len;i++) {
				tem = argv[bottom+i];
				argv[bottom+i] = argv[middle+i];
				argv[middle+i] = tem;
				SWAP_FLAGS(bottom+i, middle+i);
			}
			bottom += len;
		}
	}
	first_nonopt += (optind - last_nonopt);
	last_nonopt = optind;
}

char *
my_index(char *str, char * chr)
{
	while(*str) {
		if(hg_ishangul(chr[0],chr[1])) {
			if((*str == *chr) && (*(str+1) == *(chr+1))) return (char *) str;
			str += 2;
		} else {
			printf("[]");
			if(*str == *chr) return (char *) str;
			str++;
		}
	}
	return 0;
}

char * _getopt_initialize(int argc, char **argv, char *optstring)
{
	first_nonopt = last_nonopt = optind;
	nextchar = NULL;

	posixly_correct = (char *)getenv("POSIXLY_CORRECT");

	if(optstring[0] == '-') {
		ordering = RETURN_IN_ORDER;
		++optstring;
	} else if(optstring[0] == '+') {
		ordering = REQUIRE_ORDER;
		++optstring;
	} else if(posixly_correct != NULL)
		ordering = REQUIRE_ORDER;
	else
		ordering = PERMUTE;
	return optstring;
}

char *
_getopt_internal(int argc, char **argv, char *optstring, struct option *longopts, int *longind, int long_only)
{
	int print_errors = opterr;
	if(optstring[0] == ':')
		print_errors = 0;
	if(argc < 1)
		return NULL;

	optarg = NULL;
	if(optind == 0 || !__getopt_initialized) {
		if(optind == 0)
			optind = 1;
		optstring = _getopt_initialize(argc, argv, optstring);
		__getopt_initialized = 1;
	}

	if(nextchar == NULL || *nextchar == '\0') {
		if(last_nonopt > optind)
			last_nonopt = optind;
		if(first_nonopt > optind)
			first_nonopt = optind;
		if(ordering == PERMUTE) {
			if(first_nonopt != last_nonopt && last_nonopt != optind)
				exchange((char **) argv);
			else if(last_nonopt != optind)
				first_nonopt = optind;
			while(optind < argc && NONOPTION_P)
				optind++;
			last_nonopt = optind;
		}
		if(optind != argc && !strcmp(argv[optind], "--")) {
			optind++;
			if(first_nonopt != last_nonopt && last_nonopt != optind)
				exchange((char **) argv);
			else if(first_nonopt == last_nonopt)
				first_nonopt = optind;
			last_nonopt = argc;

			optind = argc;
		}
		if(optind == argc) {
			if(first_nonopt != last_nonopt)
				optind = first_nonopt;
			return NULL;
		}
		if(NONOPTION_P) {
			if(ordering == REQUIRE_ORDER)
				return NULL;
			optarg = argv[optind++];
			return NULL;
		}
		nextchar = (argv[optind] + 1
			    + (longopts != NULL && argv[optind][1] == '-'));
	}
	if(longopts != NULL
	   && (argv[optind][1] == '-'
	   || (long_only && (argv[optind][2]
	   || !my_index(optstring, &(argv[optind][1])))))) {
		char *nameend;
		struct option *p;
		struct option *pfound = NULL;
		int exact = 0;
		int ambig = 0;
		int indfound = -1;
		int option_index;
		char * c = (char *)malloc(2);

		for(nameend = nextchar; *nameend && *nameend != '='; nameend++)
			;

		for(p = longopts, option_index = 0; p->name; p++, option_index++)
			if(!strncmp(p->name, nextchar, nameend - nextchar)) {
				if((unsigned int) (nameend - nextchar)
				   == (unsigned int) strlen(p->name)) {
					pfound = p;
					indfound = option_index;
					exact = 1;
					break;
				} else if(pfound == NULL) {
					pfound = p;
					indfound = option_index;
				} else if(long_only
					  || pfound->has_arg != p->has_arg
					  || pfound->flag != p->flag
					  || pfound->val != p->val)
					ambig = 1;
			}
		if(ambig && !exact) {
			if(print_errors)
				fprintf(stderr, "%s: option `%s' is ambiguous\n", argv[0], argv[optind]);
			nextchar += strlen(nextchar);
			optind++;
			optopt = 0;
			return "?";
		}
		if(pfound != NULL) {
			option_index = indfound;
			optind++;
			if(*nameend) {
				if(pfound->has_arg)
					optarg = nameend + 1;
				else {
					if(print_errors) {
						if(argv[optind-1][1] == '-')
							fprintf(stderr, "%s: option '--%s' doesn't allow an argument\n", argv[0], pfound->name);
						else
							fprintf(stderr, "%s: option '%d%s' doesn't allow an argument\n", argv[0], argv[optind-1][0], pfound->name);
					}
					nextchar += strlen(nextchar);
					optopt = pfound->val;
					return "?";
				}
			} else if(pfound->has_arg == 1) {
				if(optind < argc)
					optarg = argv[optind++];
				else {
					if(print_errors)
						fprintf(stderr, "%s: option '%s' requires an argument\n", argv[0], argv[optind-1]);
					nextchar += strlen(nextchar);
					optopt = pfound->val;
					return optstring[0] == ':' ? ":" : "?";
				} 
			}
			nextchar += strlen(nextchar);
			if(longind != NULL)
				*longind = option_index;
			if(pfound->flag) {
				*(pfound->flag) = pfound->val;
				return NULL;
			}
			c[0] = pfound->val;
			c[1] = '\0';
			return c;
		}
		if(!long_only || argv[optind][1] == '-'
		   || my_index(optstring, nextchar) == NULL) {
			if(print_errors) {
				if(argv[optind][1] == '-')
					fprintf(stderr, "%s: unrecognized option '--%s'\n", argv[0], nextchar);
				else
					fprintf(stderr, "%s: unrecognized option '%c%s'\n", argv[0], argv[optind][0], nextchar);

			}
			nextchar = (char *) "";
			optind++;
			optopt = 0;
			return "?";
		}
	}
	{
		char * c;
		char *temp;

		c = (char *) malloc(2);
		/* patched by aso */
		if(hg_ishangul(*nextchar, *(nextchar+1))) {
			c[0] = *nextchar++;
			c[1] = *nextchar++;
		} else {
			c[0] = *nextchar++;
		}
		temp = my_index(optstring, c);

		if(*nextchar == '\0')
			++optind;
		if(temp == NULL || c[0] == ':') {
			if(print_errors) {
				if(posixly_correct)
					fprintf(stderr, "%s: illegal option -- %s\n", argv[0], c);
				else
					fprintf(stderr, "%s: invalid option -- %s\n", argv[0], c);
			}
			optopt = c[0];
			return "?";
		}
		if(temp[0] == 'W' && temp[1] == ';') {
			char *nameend;
			struct option *p;
			struct option *pfound = NULL;
			int exact = 0;
			int ambig = 0;
			int indfound = 0;
			int option_index;

			if(*nextchar != '\0') {
				optarg = nextchar;
				optind++;
			} else if(optind == argc) {
				if(print_errors)
					fprintf(stderr, "%s: option requires an argumnet -- %c\n", argv[0], c[0]);
				optopt = c[0];
				if(optstring[0] == ':')
					c[0] = ':';
				else
					c[0] = '?';
				return c;
			} else
				optarg = argv[optind++];
			for(nextchar = nameend = optarg; *nameend && *nameend != '='; nameend++)
				;
			for(p = longopts, option_index = 0; p->name; p++, option_index++)
				if(!strncmp(p->name, nextchar, nameend-nextchar)) {
					if((unsigned int) (nameend - nextchar) == strlen(p->name)) {
						pfound = p;
						indfound = option_index;
						exact = 1;
						break;
					} else if(pfound == NULL) {
						pfound = p;
						indfound = option_index;
					} else
						ambig = 1;
				}
			if(ambig && !exact) {
				if(print_errors)
					fprintf(stderr, "%s: option '-W %s' is ambiguous\n", argv[0], argv[optind]);
				nextchar += strlen(nextchar);
				optind++;
				return "?";
			}
			if(pfound != NULL) {
				option_index = indfound;
				if(*nameend) {
					if(pfound->has_arg)
						optarg = nameend+1;
					else {
						if(print_errors)
							fprintf(stderr, "%s: option '-W %s' doesn't allow an argument\n", argv[0], pfound->name);
						nextchar += strlen(nextchar);
						return "?";
					}
				} else if(pfound->has_arg == 1) {
					if(optind < argc)
						optarg = argv[optind++];
					else {
						if(print_errors)
							fprintf(stderr, "%s: option '%s' requires an argument\n", argv[0], argv[optind-1]);
						nextchar += strlen(nextchar);
						return optstring[0] == ':' ? ":" : "?";
					}
				}
				nextchar += strlen(nextchar);
				if(longind != NULL)
					*longind = option_index;
				if(pfound->flag) {
					*(pfound->flag) = pfound->val;
					return NULL;
				}
				c[0] = pfound->val;
				c[1] = '\0';
				return c;
			}
			nextchar = NULL;
			return "W";
		}
		if(temp[1] == ':' || (hg_ishangul(c[0], c[1]) && temp[2] == ':') ) {
			if(temp[2] == ':' && !(hg_ishangul(c[0], c[1])) ) {
				if(*nextchar != '\0') {
					optarg = nextchar;
					optind++;
				} else
					optarg = NULL;
				nextchar = NULL;
			} else {
				if(*nextchar != '\0') {
					optarg = nextchar;
					optind++;
				} else if(optind == argc) {
					if(print_errors)
						fprintf(stderr, "%s: option requires an argument -- %c\n", argv[0], c[0]);
					optopt = c[0];
					if(optstring[0] == ':')
						c[0] = ':';
					else
						c[0] = '?';
				} else
					optarg = argv[optind++];
				nextchar = NULL;
			}
		}
		return c;
	}
}

char *
getopt_long(int argc, char **argv, char *options, struct option *long_options, int *opt_index)
{
	return _getopt_internal(argc, argv, options, long_options, opt_index, 0);
}

char *
getopt_long_only(int argc, char **argv, char *options, struct option *long_options, int *opt_index)
{
	return _getopt_internal(argc, argv, options, long_options, opt_index, 1);
}

int main(int argc, char *argv[])
{
	char * c;

	while(1) {
		int option_index=0;
		static struct option long_options[] =
		{
			{ "add",     1, 0, 0 },
			{ "append",  0, 0, 0 },
			{ "delete",  1, 0, 0 },
			{ "verbose", 0, 0, 0 },
			{ "create",  0, 0, 0 },
			{ "file",    1, 0, 0 },
			{ "߰",    1, 0, 0 },
			{ 0,         0, 0, 0 }
		};

		c = getopt_long(argc, argv, ":bc:d:0123456789:", long_options, &option_index);
		if(c == NULL)
			break;
		printf("after = %s %s\n", c, optarg);
	}

	if(optind < argc) {
		printf("non-option ARGV-elements: ");
		while(optind < argc)
			printf("%s ", argv[optind++]);
		printf("\n");
	}

	exit(0);
}
